use crate::prelude::{DateTime, Id, Link, Mailbox, Recipients};
use std::fmt;

#[derive(Debug)]
pub struct PreBlk<'a> {
    alt: Option<&'a str>,
    lines: Vec<&'a str>,
}

#[derive(Debug, Default)]
enum State<'a> {
    #[default]
    Normal,
    Preformatted(PreBlk<'a>),
    Quote(Vec<&'a str>),
}

#[derive(Clone, Debug, PartialEq)]
pub enum GemtextNode {
    Sender(Mailbox),
    Recipients(Recipients),
    Timestamp(DateTime),
    Text(String),
    Heading1(String),
    Heading2(String),
    Heading3(String),
    ListItem(String),
    Quote(String),
    Preformatted(Option<String>, String),
    Link(Link),
}

impl From<Id> for GemtextNode {
    fn from(value: Id) -> Self {
        Self::Timestamp(value.into())
    }
}

impl fmt::Display for GemtextNode {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::Recipients(r) => writeln!(f, "{r}"),
            Self::Sender(m) => writeln!(f, "{m}"),
            Self::Timestamp(t) => writeln!(f, "@ {t}"),
            Self::Text(t) => writeln!(f, "{t}"),
            Self::Heading1(h) => writeln!(f, "# {h}"),
            Self::Heading2(h) => writeln!(f, "## {h}"),
            Self::Heading3(h) => writeln!(f, "### {h}"),
            Self::ListItem(l) => writeln!(f, "* {l}"),
            Self::Quote(q) => writeln!(f, "> {q}"),
            Self::Preformatted(a, p) => match a {
                None => writeln!(f, "```\n{p}\n```"),
                Some(alt) => writeln!(f, "```{alt}\n{p}\n```"),
            },
            Self::Link(l) => writeln!(f, "=> {l}"),
        }
    }
}

impl<'a> GemtextNode {
    fn parse_link(text: &'a str) -> Self {
        if let Ok(link) = text.parse() {
            Self::Link(link)
        } else {
            Self::Text(text.to_string())
        }
    }

    fn parse_heading(text: &'a str) -> Self {
        if let Some((h, s)) = text.split_once(char::is_whitespace) {
            match h {
                "#" => Self::Heading1(s.to_string()),
                "##" => Self::Heading2(s.to_string()),
                "###" => Self::Heading3(s.to_string()),
                _ => Self::Text(text.to_string()),
            }
        } else {
            Self::Text(text.to_string())
        }
    }

    fn parse_list_item(text: &'a str) -> Self {
        match text.split_once(char::is_whitespace) {
            Some((pre, s)) if pre == "*" => GemtextNode::ListItem(s.to_string()),
            _ => GemtextNode::Text(text.to_string()),
        }
    }

    fn parse_senders(text: &'a str) -> Self {
        let Some(line) = text.strip_prefix('<').map(|x| x.trim()) else {
            return Self::Text(text.to_string());
        };
        if let Ok(user) = line.parse() {
            Self::Sender(user)
        } else {
            Self::Text(text.to_string())
        }
    }

    fn parse_recipients(text: &'a str) -> Self {
        let Some(line) = text.strip_prefix(':') else {
            return Self::Text(text.to_string());
        };
        let split = line.split_whitespace();
        let mut recipients: Recipients = Recipients { boxes: vec![] };
        for s in split {
            if let Ok(m) = s.parse() {
                recipients.boxes.push(m);
            } else {
                return Self::Text(text.to_string());
            }
        }
        Self::Recipients(recipients)
    }

    fn parse_timestamp(text: &'a str) -> Self {
        let Some(line) = text.strip_prefix('@').map(str::trim) else {
            return Self::Text(text.to_string());
        };
        if let Ok(dt) = line.parse() {
            Self::Timestamp(dt)
        } else {
            Self::Text(text.to_string())
        }
    }
}

#[derive(Debug, Default)]
pub struct Parser<'a> {
    state: State<'a>,
    title: Option<String>,
    lines: Vec<GemtextNode>,
}

impl<'a> Parser<'a> {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn parse(mut self, raw: &'a str) -> Vec<GemtextNode> {
        for line in raw.lines() {
            match self.state {
                State::Normal => self.parse_normal(line),
                State::Preformatted(_) => self.parse_preformatted(line),
                State::Quote(_) => self.parse_quote(line),
            }
        }
        match self.state {
            State::Normal => {}
            State::Preformatted(_) => self.leave_preformatted(),
            State::Quote(q) => {
                let quote = q.join("\n");
                self.lines.push(GemtextNode::Quote(quote));
            }
        }
        self.lines
    }

    fn link(&mut self, line: &'a str) {
        self.lines.push(GemtextNode::parse_link(line));
    }

    fn heading(&mut self, line: &'a str) {
        let line = GemtextNode::parse_heading(line);
        if self.title.is_none() {
            if let GemtextNode::Heading1(t) = &line {
                self.title = Some(t.clone());
            }
        }
        self.lines.push(line);
    }

    fn list_item(&mut self, line: &'a str) {
        self.lines.push(GemtextNode::parse_list_item(line));
    }

    fn enter_quote(&mut self, line: &'a str) {
        match line.split_once(char::is_whitespace) {
            Some((prefix, suffix)) if prefix == ">" => self.state = State::Quote(vec![suffix]),
            _ => self.lines.push(GemtextNode::Text(line.to_string())),
        }
    }

    fn leave_quote(&mut self, line: &'a str) {
        match &mut self.state {
            State::Quote(q) => {
                let quote = q.join("\n");
                self.lines.push(GemtextNode::Quote(quote));
            }
            _ => panic!("Attempt to parse as quote when not in quote mode"),
        }
        self.state = State::Normal;
        self.lines.push(GemtextNode::Text(line.to_string()));
    }

    fn enter_preformatted(&mut self, line: &'a str) {
        let alt = if line.len() > 3 {
            Some(line[3..].trim())
        } else {
            None
        };
        let preblk = PreBlk { alt, lines: vec![] };
        self.state = State::Preformatted(preblk);
    }

    fn leave_preformatted(&mut self) {
        match &self.state {
            State::Preformatted(v) => {
                let s = v.lines.join("\n");
                self.lines
                    .push(GemtextNode::Preformatted(v.alt.map(str::to_string), s));
                self.state = State::Normal;
            }
            _ => panic!("Attempted to leave preformatted mode when not in preformatted mode"),
        }
    }

    fn senders(&mut self, line: &'a str) {
        self.lines.push(GemtextNode::parse_senders(line));
    }

    fn recipients(&mut self, line: &'a str) {
        self.lines.push(GemtextNode::parse_recipients(line));
    }

    fn timestamp(&mut self, line: &'a str) {
        self.lines.push(GemtextNode::parse_timestamp(line));
    }

    fn parse_normal(&mut self, line: &'a str) {
        match line {
            s if s.starts_with("=>") => self.link(s),
            s if s.starts_with('#') => self.heading(s),
            s if s.starts_with('*') => self.list_item(s),
            s if s.starts_with('>') => self.enter_quote(s),
            s if s.starts_with("```") => self.enter_preformatted(s),
            s if s.starts_with('<') => self.senders(s),
            s if s.starts_with(':') => self.recipients(s),
            s if s.starts_with('@') => self.timestamp(s),
            s => self.lines.push(GemtextNode::Text(s.to_string())),
        }
    }

    fn parse_preformatted(&mut self, line: &'a str) {
        if line.starts_with("```") {
            self.leave_preformatted();
        } else {
            match &mut self.state {
                State::Preformatted(p) => p.lines.push(line),
                _ => panic!("Attempt to parse as preformatted when not in preformatted mode"),
            }
        }
    }

    fn parse_quote(&mut self, line: &'a str) {
        if let Some(suffix) = line.strip_prefix('>') {
            match &mut self.state {
                State::Quote(q) => q.push(suffix.trim()),
                _ => panic!("Attempt to parse as quote when not in quote mode"),
            }
        } else {
            self.leave_quote(line);
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::prelude::Host;

    use super::*;

    static RAW: &'static str =
        include_str!("../../test/mailstore/misfin.example.org/jane/Lists/700070.gmi");

    fn parse_raw() -> Vec<GemtextNode> {
        Parser::new().parse(RAW)
    }

    #[test]
    fn nodes() {
        let nodes = parse_raw();
        let mut nodes = nodes.iter();
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::Sender(Mailbox {
                username: "ed".to_string(),
                host: Host {
                    subdomain: None,
                    domain: "iron".to_string(),
                    tld: "maiden".to_string()
                },
                blurb: Some("Eddy the Head".to_string())
            })
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::Sender(Mailbox {
                username: "bruce".to_string(),
                host: Host {
                    subdomain: None,
                    domain: "dickinson".to_string(),
                    tld: "jam".to_string(),
                },
                blurb: Some("Bruce Dickinson".to_string()),
            })
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::Recipients(Recipients {
                boxes: vec![
                    Mailbox {
                        username: "nicko".to_string(),
                        host: Host {
                            subdomain: None,
                            domain: "iron".to_string(),
                            tld: "maiden".to_string(),
                        },
                        blurb: None,
                    },
                    Mailbox {
                        username: "steve".to_string(),
                        host: Host {
                            subdomain: None,
                            domain: "iron".to_string(),
                            tld: "maiden".to_string()
                        },
                        blurb: None,
                    }
                ]
            })
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::Timestamp("2023-06-07T21:27:15Z".parse().unwrap()),
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::Text("".to_string()),
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::Heading1("Checking in".to_string()),
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::Text("The new album's in the can. A few things:".to_string()),
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::ListItem("Needs more cowbell".to_string()),
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::ListItem(
                "Adrian's guitar is too loud, drowning out the cowbell".to_string()
            ),
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::ListItem("Steve has a fever".to_string()),
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::Link(Link {
                url: "gemini://private.iron.maiden/clip.mp3".to_string(),
                display: Some("clip".to_string())
            })
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::Text("".to_string()),
        );
        assert_eq!(
            nodes.next().cloned().unwrap(),
            GemtextNode::Quote("You're gonna want that cowbell!".to_string()),
        );
        assert!(nodes.next().is_none());
    }
}
