use {
    super::{Message, Recipients},
    crate::prelude::Mailbox,
};

#[derive(Debug, Default)]
pub struct Parser {
    id: String,
    from: Option<Mailbox>,
    senders: Vec<Mailbox>,
    recipients: Recipients,
    timestamp: Option<String>,
    title: Option<String>,
    body: String,
}

impl Parser {
    pub fn new(id: &str) -> Self {
        Self {
            id: id.to_string(),
            ..Default::default()
        }
    }

    #[allow(clippy::missing_panics_doc)]
    pub fn parse(mut self, content: &str) -> Result<Message, super::Error> {
        let lines = content.lines();
        for l in lines {
            match l {
                s if s.starts_with('<') && self.from.is_none() && self.body.is_empty() => {
                    let s = s.strip_prefix('<').unwrap().trim();
                    let from: Mailbox = s.parse()?;
                    self.from = Some(from);
                }
                s if s.starts_with('<') && self.body.is_empty() => {
                    let sndr = s.strip_prefix('<').unwrap().trim();
                    let from: Mailbox = sndr.parse()?;
                    self.senders.push(from);
                }
                s if s.starts_with(':') && self.body.is_empty() => {
                    self.recipients = s.parse()?;
                }
                s if s.starts_with('@') && self.timestamp.is_none() && self.body.is_empty() => {
                    self.timestamp = Some(s.strip_prefix('@').unwrap().trim().to_string());
                }
                s if s.starts_with("###") && self.title.is_none() => {
                    if let Some(t) = s.strip_prefix("###").map(|x| x.trim().to_string()) {
                        self.title = Some(t);
                    }
                    if !self.body.is_empty() {
                        self.body.push('\n');
                    }
                    self.body.push_str(s);
                }
                s if s.starts_with("##") && self.title.is_none() => {
                    if let Some(t) = s.strip_prefix("##").map(|x| x.trim().to_string()) {
                        self.title = Some(t);
                    }
                    if !self.body.is_empty() {
                        self.body.push('\n');
                    }
                    self.body.push_str(s);
                }
                s if s.starts_with('#') && self.title.is_none() => {
                    if let Some(t) = s.strip_prefix('#').map(|x| x.trim().to_string()) {
                        self.title = Some(t);
                    }
                    if !self.body.is_empty() {
                        self.body.push('\n');
                    }
                    self.body.push_str(s);
                }
                s => {
                    if !self.body.is_empty() {
                        self.body.push('\n');
                    }
                    self.body.push_str(s);
                }
            }
        }
        if self.from.is_none() {
            Err(super::Error::EmptySender)
        } else {
            Ok(Message {
                id: self.id,
                from: self.from.unwrap(),
                senders: self.senders,
                recipients: self.recipients.boxes,
                timestamp: self.timestamp,
                title: self.title,
                body: self.body,
            })
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    const RAW: &'static str =
        include_str!("../../test/mailstore/misfin.example.org/dick/Inbox/619310.gmi");

    #[test]
    fn parse_message() {
        let msg: Message = Parser::new("619310").parse(RAW).unwrap();
        assert_eq!(msg.id, "619310");
        assert_eq!(msg.from.to_string(), "joe@gemini.example.org");
        assert_eq!(msg.timestamp.unwrap(), "2023-06-07T16:09:42Z");
        assert!(msg.senders.is_empty());
        assert!(msg.recipients.is_empty());
        assert_eq!(msg.title.unwrap(), "How 'bout dose Bears?");
        assert_eq!(
            msg.body,
            "# How 'bout dose Bears?\nWhen are they coming for dinner anyway?\n"
        );
    }

    #[test]
    fn print_message() {
        let msg: Message = Parser::new("ox42sc69").parse(RAW).unwrap();
        assert_eq!(
            msg.to_string(),
            "< joe@gemini.example.org\n@ 2023-06-07T16:09:42Z\n# How 'bout dose Bears?\nWhen are they coming for dinner anyway?\n\r\n"
        );
    }
}
